@@ -77,37 +77,38 @@ func Copy(c *cli.Context) (err error) {
 	var vfile VPFile
 	switch parseCLIOptions(c) {
 	case copyin:
-		r, err = os.OpenFile(c.Args().First(), os.O_CREATE|os.O_RDONLY, 0655)
+		r, err = os.Open(c.Args().First())
 		defer r.Close()
 		if err != nil {
-			return cli.NewExitError(fmt.Sprintf("%v", err), 1)
+			return cli.NewExitError(err, 1)
 		}
 		vfile, err = virtualPartitionFile.Open(comp, c.Args().Get(1))
 		defer wclose(vfile)
 		if err != nil {
-			return cli.NewExitError(fmt.Sprintf("%v", err), 1)
+			return cli.NewExitError(err, 1)
 		}
-		w = vfile
+		if err = vfile.CopyTo(c.Args().First()); err != nil {
+			return cli.NewExitError(err, 1)
+		}
+		return nil
 	case copyinstdin:
 		r = os.Stdin
-		vfile, err = virtualPartitionFile.Open(comp, c.Args().First())
+		vfile, err = virtualPartitionFile.Open(comp, c.Args().Get(1))
 		defer wclose(vfile)
 		if err != nil {
-			return cli.NewExitError(fmt.Sprintf("%v", err), 1)
+			return cli.NewExitError(err, 1)
 		}
 		w = vfile
 	case copyout:
 		vfile, err = virtualPartitionFile.Open(comp, c.Args().First())
 		defer wclose(vfile)
 		if err != nil {
+			return cli.NewExitError(err, 1)
+		}
+		if err = vfile.CopyFrom(c.Args().Get(1)); err != nil {
 			return cli.NewExitError(fmt.Sprintf("%v", err), 1)
 		}
-		r = vfile
-		w, err = os.OpenFile(c.Args().Get(1), os.O_CREATE|os.O_WRONLY, 0655)
-		defer w.Close()
-		if err != nil {
-			return cli.NewExitError(fmt.Sprintf("%v", err), 1)
-		}
+		return nil
 	case parseError:
 		return cli.NewExitError(fmt.Sprintln("no artifact or sdimage provided"), 1)
 	case argerror:
@@ -117,8 +118,10 @@ func Copy(c *cli.Context) (err error) {
 	}
 
 	_, err = io.Copy(w, r)
-
-	return err
+	if err != nil {
+		return cli.NewExitError(err, 1)
+	}
+	return nil
 }
 
 // Install installs a file from the host filesystem onto either
@@ -211,33 +214,28 @@ const (
 
 func parseCLIOptions(c *cli.Context) int {
 
-	finfo, err := os.Stdin.Stat()
-	if err != nil {
-		return criterror
+	if c.NArg() != 2 {
+		return argerror
 	}
 
-	// no data on stdin
-	if finfo.Mode()&os.ModeNamedPipe == 0 {
-		if c.NArg() != 2 {
-			return argerror
-		}
-		switch {
-
-		case isimg.MatchString(c.Args().First()):
-			return copyout
-
-		case isimg.MatchString(c.Args().Get(1)):
-			return copyin
-
-		default:
+	// If the first argument is '-', read from stdin
+	if c.Args().First() == "-" {
+		// Read from stdin
+		if !isimg.MatchString(c.Args().Get(1)) {
 			return parseError
+		}
+		return copyinstdin
+	}
 
-		}
-	} else {
-		// data on stdin
-		if isimg.MatchString(c.Args().First()) {
-			return copyinstdin
-		}
+	switch {
+
+	case isimg.MatchString(c.Args().First()):
+		return copyout
+
+	case isimg.MatchString(c.Args().Get(1)):
+		return copyin
+
+	default:
 		return parseError
 	}
 }
